import math
import numpy as np
from nibabel.eulerangles import euler2mat
from numpy.fft import fftn, ifftn, fftshift, ifftshift

# Regularization scaling
# For weighted filtering, the regularization term should be a fraction of the noise level
# we use 0.01 by default
REG_SCALE = 0.01


def _get_dim(data):
    """
    Data dimensionality with error checking
    """

    if data.ndim == 2:
        ndim = 2
    elif data.ndim == 3:
        ndim = 3
    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                data.ndim
            )
        )

    return ndim


def _pad(d, scale):
    """Pad for gaussian convolutions in the fourier domain

    Only need to pad by a quarter width on either side which
    gives 1/2 a kernel width when wrapped.
    """
    p = 2 * 2 ** scale
    pd = np.pad(d, pad_width=p, mode="constant", constant_values=0)
    return pd


def _crop(d, scale):
    """Inverse of pad"""
    p = 2 * 2 ** scale
    if d.ndim == 2:
        return d[p:-p, p:-p]
    elif d.ndim == 3:
        return d[p:-p, p:-p, p:-p]


def _scale_coordinates(shape, scale):
    """
    Compute the scaled image coordinates on the box [-pi,pi]^d
    """
    ndim = len(shape)
    # Compute the scaled coordinate system
    if ndim == 2:
        nx, ny = shape
        fx, fy = np.meshgrid(
            np.linspace(-nx / 2, nx / 2 - 1, nx),
            np.linspace(-ny / 2, ny / 2 - 1, ny),
            indexing="ij",
        )
        sx = nx / (2.0 * math.pi) / 2 ** scale
        sy = ny / (2.0 * math.pi) / 2 ** scale
        x = fx / sx
        y = fy / sy
        return x, y

    elif ndim == 3:
        nx, ny, nz = shape
        fx, fy, fz = np.meshgrid(
            np.linspace(-nx / 2, nx / 2 - 1, nx),
            np.linspace(-ny / 2, ny / 2 - 1, ny),
            np.linspace(-nz / 2, nz / 2 - 1, nz),
            indexing="ij",
        )
        sx = nx / (2.0 * math.pi) / 2 ** scale
        sy = ny / (2.0 * math.pi) / 2 ** scale
        sz = nz / (2.0 * math.pi) / 2 ** scale
        x = fx / sx
        y = fy / sy
        z = fz / sz

        return x, y, z

    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                len(shape)
            )
        )


def _radius(shape, scale):
    """
    Compute the radius in scaled image coordinates
    """
    x = _scale_coordinates(shape, scale)
    r = np.sqrt(np.sum(np.square(x), axis=0))
    return r


def _crop_filter(input_filter, scale, radius=None):
    """
    Crop an image domain filter
    """
    ndim = input_filter.ndim

    if radius is None:
        radius = 4 * 2 ** scale

    if ndim == 2:
        nx, ny = input_filter.shape
        output_filter = input_filter[
            int(nx / 2 - radius) : int(nx / 2 + radius + 1),
            int(ny / 2 - radius) : int(ny / 2 + radius + 1),
        ]
    elif ndim == 3:
        nx, ny, nz = input_filter.shape
        output_filter = input_filter[
            int(nx / 2 - radius) : int(nx / 2 + radius + 1),
            int(ny / 2 - radius) : int(ny / 2 + radius + 1),
            int(nz / 2 - radius) : int(nz / 2 + radius + 1),
        ]
    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                ndim
            )
        )

    return output_filter


def _pinv(x, sigma):
    """
    Pseudoinverse with regularization
    sigma is the noise level
    pinv(x) = x / (x^2 + sigma^2)

    """
    xinv = x / (x ** 2 + sigma ** 2)
    return xinv


def gaussian(data, scale=1):
    """
    Rotationally symmetric Gaussian filter in the fourier domain

    Parametrized by a scale, resolution reduced by 2^(scale+1).
    Assumes the pixels are all the same size.
    Pad and crop to prevent wrap around junk.
    """
    pd = _pad(data, scale)

    # Get the radius scaled coordinate system
    r = _radius(pd.shape, scale)

    # Compute filter as a function of radius
    g = np.exp(-0.5 * r ** 2)

    # Apply the filter
    temp = ifftn(ifftshift(g * fftshift(fftn(pd))))

    # Crop
    output = _crop(temp, scale)

    # Ensure that real functions stay real
    if np.isrealobj(data):
        return np.real(output)
    else:
        return output


def gradient(data, scale=1):
    """
    Gradient, Gaussian 1st order partial derivative filter in the fourier domain
    """

    # Gausian gradient in each direction
    # i*x*g, i*y*g, i*z*g etc.

    # Pad
    pd = _pad(data, scale)

    # Get the scaled coordinate system
    if data.ndim == 2:
        x, y = _scale_coordinates(pd.shape, scale)
        rsq = x ** 2 + y ** 2
        g = np.exp(-0.5 * rsq)
        temp = 1j * g * fftshift(fftn(pd))
        dx = ifftn(ifftshift(x * temp))
        dy = ifftn(ifftshift(y * temp))
        # Crop
        dx = _crop(dx, scale)
        dy = _crop(dy, scale)
        # Ensure that real functions stay real
        if np.isrealobj(data):
            dx = np.real(dx)
            dy = np.real(dy)
        return [dx, dy]

    elif data.ndim == 3:
        x, y, z = _scale_coordinates(pd.shape, scale)
        rsq = x ** 2 + y ** 2 + z ** 2
        g = np.exp(-0.5 * rsq)
        temp = 1j * g * fftshift(fftn(pd))
        dx = ifftn(ifftshift(x * temp))
        dy = ifftn(ifftshift(y * temp))
        dz = ifftn(ifftshift(z * temp))
        # Crop
        dx = _crop(dx, scale)
        dy = _crop(dy, scale)
        dz = _crop(dz, scale)
        # Ensure that real functions stay real
        if np.isrealobj(data):
            dx = np.real(dx)
            dy = np.real(dy)
            dz = np.real(dz)
        return [dx, dy, dz]
    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                data.ndim
            )
        )


def hessian(data, scale=1):
    """
    Hessian, Gaussian 2nd order partial derivatives filter in the fourier domain
    """

    # Gausian 2nd derivative in each direction
    # (i*x)*(i*y)*g, etc

    # Pad
    pd = _pad(data, scale)

    # Get the scaled coordinate system
    if data.ndim == 2:
        x, y = _scale_coordinates(pd.shape, scale)
        rsq = x ** 2 + y ** 2
        g = np.exp(-0.5 * rsq)
        temp = -1.0 * g * fftshift(fftn(pd))
        dxx = ifftn(ifftshift(x * x * temp))
        dxy = ifftn(ifftshift(x * y * temp))
        dyy = ifftn(ifftshift(y * y * temp))
        # Crop
        dxx = _crop(dxx, scale)
        dxy = _crop(dxy, scale)
        dyy = _crop(dyy, scale)
        # Ensure that real functions stay real
        if np.isrealobj(data):
            dxx = np.real(dxx)
            dxy = np.real(dxy)
            dyy = np.real(dyy)
        return [dxx, dxy, dyy]

    elif data.ndim == 3:
        x, y, z = _scale_coordinates(pd.shape, scale)
        rsq = x ** 2 + y ** 2 + z ** 2
        g = np.exp(-0.5 * rsq)
        temp = -1.0 * g * fftshift(fftn(pd))
        dxx = ifftn(ifftshift(x * x * temp))
        dxy = ifftn(ifftshift(x * y * temp))
        dxz = ifftn(ifftshift(x * z * temp))
        dyy = ifftn(ifftshift(y * y * temp))
        dyz = ifftn(ifftshift(y * z * temp))
        dzz = ifftn(ifftshift(z * z * temp))
        # Crop
        dxx = _crop(dxx, scale)
        dxy = _crop(dxy, scale)
        dxz = _crop(dxz, scale)
        dyy = _crop(dyy, scale)
        dyz = _crop(dyz, scale)
        dzz = _crop(dzz, scale)
        # Ensure that real functions stay real
        if np.isrealobj(data):
            dxx = np.real(dxx)
            dxy = np.real(dxy)
            dxz = np.real(dxz)
            dyy = np.real(dyy)
            dyz = np.real(dyz)
            dzz = np.real(dzz)
        return [dxx, dxy, dxz, dyy, dyz, dzz]

    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                data.ndim
            )
        )


def high_pass(data, scale=1):
    """
    High pass filter
    data - G(data,s)
    """
    hp = data - gaussian(data, scale=scale)

    return hp


def gaussian_kernel(shape, scale=1):
    """
    Rotationally symmetric Gaussian filter kernel in the image domain
    """

    # Get the radius scaled coordinate system
    r = _radius(shape, scale)

    # Compute filter as a function of radius
    g = np.exp(-0.5 * r ** 2)

    # The filter in the image domain
    im_filter = np.real(fftshift(ifftn(fftshift(g))))

    # Cropped
    im_filter = _crop_filter(im_filter, scale)

    return im_filter


def gradient_kernel(shape, scale=1):
    """
    Gradient, Gaussian 1st order partial derivative filter kernel in the image domain
    """

    # Gausian gradient in each direction
    # i*x*g, i*y*g, i*z*g etc.
    ndim = len(shape)

    if ndim == 2:
        x, y = _scale_coordinates(shape, scale)
        rsq = x ** 2 + y ** 2
        g = np.exp(-0.5 * rsq)
        dx = np.real(fftshift(ifftn(ifftshift(1j * x * g))))
        dy = np.real(fftshift(ifftn(ifftshift(1j * y * g))))
        dx = _crop_filter(dx, scale)
        dy = _crop_filter(dy, scale)
        return [dx, dy]

    elif ndim == 3:
        x, y, z = _scale_coordinates(shape, scale)
        rsq = x ** 2 + y ** 2 + z ** 2
        g = np.exp(-0.5 * rsq)
        dx = np.real(fftshift(ifftn(ifftshift(1j * x * g))))
        dy = np.real(fftshift(ifftn(ifftshift(1j * y * g))))
        dz = np.real(fftshift(ifftn(ifftshift(1j * z * g))))
        dx = _crop_filter(dx, scale)
        dy = _crop_filter(dy, scale)
        dz = _crop_filter(dz, scale)
        return [dx, dy, dz]
    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                ndim
            )
        )


def hessian_kernel(shape, scale=1):
    """
    Hessian, Gaussian 2nd order partial derivatives filter kernel in the image domain
    """

    # Gausian 2nd derivative in each direction
    # (i*x)*(i*y)*g, etc
    ndim = len(shape)

    if ndim == 2:
        x, y = _scale_coordinates(shape, scale)
        rsq = x ** 2 + y ** 2
        g = np.exp(-0.5 * rsq)
        dxx = np.real(fftshift(ifftn(ifftshift(-1.0 * x * x * g))))
        dxy = np.real(fftshift(ifftn(ifftshift(-1.0 * x * y * g))))
        dyy = np.real(fftshift(ifftn(ifftshift(-1.0 * y * y * g))))
        dxx = _crop_filter(dxx, scale)
        dxy = _crop_filter(dxy, scale)
        dyy = _crop_filter(dyy, scale)
        return [dxx, dxy, dyy]

    elif ndim == 3:
        x, y, z = _scale_coordinates(shape, scale)
        rsq = x ** 2 + y ** 2 + z ** 2
        g = np.exp(-0.5 * rsq)
        dxx = np.real(fftshift(ifftn(ifftshift(-1.0 * x * x * g))))
        dxy = np.real(fftshift(ifftn(ifftshift(-1.0 * x * y * g))))
        dxz = np.real(fftshift(ifftn(ifftshift(-1.0 * x * z * g))))
        dyy = np.real(fftshift(ifftn(ifftshift(-1.0 * y * y * g))))
        dyz = np.real(fftshift(ifftn(ifftshift(-1.0 * y * z * g))))
        dzz = np.real(fftshift(ifftn(ifftshift(-1.0 * z * z * g))))
        dxx = _crop_filter(dxx, scale + 1)
        dxy = _crop_filter(dxy, scale + 1)
        dxz = _crop_filter(dxz, scale + 1)
        dyy = _crop_filter(dyy, scale + 1)
        dyz = _crop_filter(dyz, scale + 1)
        dzz = _crop_filter(dzz, scale + 1)
        return [dxx, dxy, dxz, dyy, dyz, dzz]

    else:
        raise RuntimeError(
            "Unsupported number of dimensions {}. We only supports 2 or 3D arrays.".format(
                ndim
            )
        )


def gradient_amplitude(g):
    """
    Amplitude in the gradient filter band
    2/Frobenius norm
    """
    amp = np.sqrt(np.sum(np.abs(g) ** 2, axis=0))
    return amp


def hessian_amplitude(h):
    """
    Amplitude in the hessian filter band
    Frobenius norm
    """
    if len(h) == 3:
        a = np.sqrt(np.abs(h[0]) ** 2 + 2 * np.abs(h[1]) ** 2 + np.abs(h[2]) ** 2)
    elif len(h) == 6:
        a = np.sqrt(
            np.abs(h[0]) ** 2
            + 2 * np.abs(h[1]) ** 2
            + 2 * np.abs(h[2]) ** 2
            + np.abs(h[3]) ** 2
            + 2 * np.abs(h[4]) ** 2
            + np.abs(h[5]) ** 2
        )
    else:
        raise RuntimeError("Unsupported number of dimensions {}.".format(len(h)))
    return a


def hessian_det(h):
    """
    Determinant in the hessian filter band
    """
    if len(h) == 3:
        det = h[0] * h[2] - h[1] * h[1]
    elif len(h) == 6:
        det = (
            h[0] * (h[3] * h[5] - h[4] * h[4])
            - h[1] * (h[1] * h[5] - h[4] * h[2])
            + h[2] * (h[1] * h[4] - h[3] * h[2])
        )
    else:
        raise RuntimeError("Unsupported number of dimensions {}.".format(len(h)))
    return det


def hessian_rot(h):
    """
    Rotational invariants of the hessian
    """

    if len(h) == 3:
        # [dxx, dxy, dyy]
        # 1st trace l1 + l2
        trace = h[0] + h[2]
        # 2nd determinant l1*l2
        # det = Dxx*Dyy - Dxy*Dyx
        det = h[0] * h[2] - h[1] * h[1]
        # frobenius norm sqrt(l1**2 + l2**2)
        frobenius = np.sqrt(
            np.abs(h[0]) ** 2 + 2 * np.abs(h[1]) ** 2 + np.abs(h[2]) ** 2
        )

        return trace, det, frobenius

    elif len(h) == 6:
        # [dxx, dxy, dxz, dyy, dyz, dzz]
        # 1st trace l1 + l2 + l3
        trace = h[0] + h[3] + h[5]
        # 2nd l1*l2 + l1*l3 + l2*l3
        # sec = Dxx*Dyy - Dxy*Dyx + Dxx*Dzz - Dxz*Dzx + Dyy*Dzz - Dyz*Dzy
        sec = (
            h[0] * h[3]
            - h[1] * h[1]
            + h[0] * h[5]
            - h[2] * h[2]
            + h[3] * h[5]
            - h[4] * h[4]
        )
        # 3rd determinant l1*l2*l3
        # det = Dxx*(Dyy*Dzz - Dyz*Dzy) - Dxy*(Dyx*Dzz - Dyz*Dzx) + Dxz*(Dyx*Dzy - Dyy*Dzx)
        det = (
            h[0] * (h[3] * h[5] - h[4] * h[4])
            - h[1] * (h[1] * h[5] - h[4] * h[2])
            + h[2] * (h[1] * h[4] - h[3] * h[2])
        )
        # frobenius norm sqrt(l1**2 + l2**2 + l3**2)
        frobenius = np.sqrt(
            np.abs(h[0]) ** 2
            + 2 * np.abs(h[1]) ** 2
            + 2 * np.abs(h[2]) ** 2
            + np.abs(h[3]) ** 2
            + 2 * np.abs(h[4]) ** 2
            + np.abs(h[5]) ** 2
        )

        return trace, sec, det, frobenius

    else:
        raise RuntimeError("Unsupported number of dimensions {}.".format(len(h)))


def hessian_trace(h):
    """
    Trace in the hessian filter band
    Laplacian
    """
    if len(h) == 3:
        trace = h[0] + h[2]
    elif len(h) == 6:
        trace = h[0] + h[3] + h[5]
    else:
        raise RuntimeError("Unsupported number of dimensions {}.".format(len(h)))
    return trace


def rotate_gradient_2d(gx, gy, z=0.0):
    """Rotate a 2d gradient or band pass gradient.

    gp = R * g
    """
    R = euler2mat(z=z)[:2, :2]
    g = np.zeros([2, gx.size], dtype=gx.dtype)
    g[0, :] = gx.ravel()
    g[1, :] = gy.ravel()
    gp = R @ g
    gxp = gp[0].reshape(gx.shape)
    gyp = gp[1].reshape(gy.shape)
    return gxp, gyp


def rotate_gradient_3d(gx, gy, gz, z=0.0, y=0.0, x=0.0):
    """Rotate a 3d gradient or band pass gradient.

    gp = R * g
    """
    R = euler2mat(z=z, y=y, x=x)
    g = np.zeros([3, gx.size], dtype=gx.dtype)
    g[0, :] = gx.ravel()
    g[1, :] = gy.ravel()
    g[2, :] = gz.ravel()
    gp = R @ g
    gxp = gp[0].reshape(gx.shape)
    gyp = gp[1].reshape(gy.shape)
    gzp = gp[2].reshape(gz.shape)
    return gxp, gyp, gzp


def rotate_hessian_2d(dxx, dxy, dyy, z):
    """Rotate a 2d hessian or band pass hessian.

    Hp = R * H * R'
    """
    R = euler2mat(z)[:2, :2]

    H = np.zeros([2, 2, dxx.size], dtype=dxx.dtype)
    H[0, 0, :] = dxx.ravel()
    H[0, 1, :] = dxy.ravel()
    H[1, 0, :] = H[0, 1, :]
    H[1, 1, :] = dyy.ravel()

    #     Hp = np.zeros(H.shape, dtype=H.dtype)
    #     for n in range(H.shape[2]):
    #         Hp[:,:,n] = R @ H[:,:,n] @ R.transpose()
    Hp = np.dot(np.dot(R, H).transpose(2, 0, 1), R.transpose()).transpose(1, 2, 0)

    dxxp = Hp[0, 0, :].reshape(dxx.shape)
    dxyp = Hp[0, 1, :].reshape(dxy.shape)
    dyyp = Hp[1, 1, :].reshape(dyy.shape)

    return dxxp, dxyp, dyyp


def rotate_hessian_3d(dxx, dxy, dxz, dyy, dyz, dzz, z, y=0.0, x=0.0):
    """Rotate a 3d hessian or band pass hessian.

    Hp = R' * H * R
    """
    R = euler2mat(z)

    H = np.zeros([3, 3, dxx.size], dtype=dxx.dtype)
    H[0, 0, :] = dxx.ravel()
    H[0, 1, :] = dxy.ravel()
    H[0, 2, :] = dxz.ravel()
    H[1, 1, :] = dyy.ravel()
    H[1, 2, :] = dyz.ravel()
    H[2, 2, :] = dzz.ravel()
    H[1, 0, :] = H[0, 1, :]
    H[2, 0, :] = H[0, 2, :]
    H[2, 1, :] = H[1, 2, :]

    #     Hp = np.zeros(H.shape, dtype=H.dtype)
    #     for n in range(H.shape[2]):
    #         Hp[:,:,n] = R.transpose() @ H[:,:,n] @ R
    Hp = np.dot(np.dot(R.transpose(), H).transpose(2, 0, 1), R).transpose(1, 2, 0)

    dxxp = Hp[0, 0, :].reshape(dxx.shape)
    dxyp = Hp[0, 1, :].reshape(dxy.shape)
    dxzp = Hp[0, 2, :].reshape(dxz.shape)
    dyyp = Hp[1, 1, :].reshape(dyy.shape)
    dyzp = Hp[1, 2, :].reshape(dyz.shape)
    dzzp = Hp[2, 2, :].reshape(dzz.shape)

    return dxxp, dxyp, dxzp, dyyp, dyzp, dzzp


def dog(d, order, scale, w=None, sigma=0.001, band_pass=False):
    """Gaussian derivative operator with regularization"""

    # Band pass
    #   call gauss twice and subtract
    if band_pass:
        a = dog(d, w, sigma, scale=scale)
        b = dog(d, w, sigma, scale=scale + 1)
        return a - b

    # Low pass
    if order == 0:
        if w is not None:
            a = gaussian(w * d, scale)
            b = _pinv(gaussian(w, scale), sigma)
            return a * b
        else:
            return gaussian(d, scale=scale)

    elif order == 1:
        if w is not None:
            a = gradient(w * d, scale)
            b = _pinv(gaussian(w, scale), sigma)
            a = [x * b for x in a]
            return a
        else:
            return gradient(d, scale=scale)

    elif order == 2:
        if w is not None:
            a = hessian(w * d, scale)
            b = _pinv(gaussian(w, scale), sigma)
            a = [x * b for x in a]
            return a
        else:
            return hessian(d, scale=scale)

    else:
        raise RuntimeError(
            (
                f"Unsupported gaussian derivative order {order}."
                f"We only supports derivatives up to order 2."
            )
        )


def dog_rotational_invariants(f, order=0):
    r"""Rotational invariants of gaussian derivative operator

    returns a list

    Principal invariants and Frobenius norm
    See https://en.wikipedia.org/wiki/Invariants_of_tensors

    Gaussian (Order 0):
        2 rotational invariant: G0, \|G0\|
    Gradient (Order 1):
        1 rotational invariant: \|G1\|
    Hessian (Order 2):
        in 2D, 3 rotational invariants: Tr(G2), Det(G2), \|G2\|
        in 3D, 4 rotational invariants: Tr(G2), Det(G2), I2(G2), \|G2\|
    """

    if order == 0:
        rot = [f, np.abs(f)]
    elif order == 1:
        rot = [gradient_amplitude(f)]
    elif order == 2:
        rot = list(hessian_rot(f))
    else:
        raise RuntimeError(
            f"Unsupported gaussian derivative order {order}."
            f"We only support derivatives of order <=2."
        )

    return rot



# -*- coding: utf-8 -*-

import click
import nibabel
import numpy as np
from .signal_stats import global_scale
from .features import (
    laplacian_pyramid,
    local_scale_normalization,
    local_contrast_normalization,
    front_end_features,
    fef_rotational_invariants,
    NUM_SCALES,
    NORMALIZATION_SCALE,
    gaussian_pyramid_features,
)


def _check_image_compatibility(images):
    im_shape = images[0].shape
    im_affine = images[0].affine
    for im in images[1:]:
        assert im.shape == im_shape, "Image shape mismatch: {} and {} differ.".format(
            im.get_filename(), images[0].get_filename()
        )
        assert np.all(
            im.affine == im_affine
        ), "Image affine mismatch: {} and {} differ.".format(
            im.get_filename(), images[0].get_filename()
        )


@click.command()
@click.option(
    "--output",
    type=click.STRING,
    default="out.nii",
    help="Output filename for the contrast normalized image.",
)
@click.option(
    "--scale", type=click.INT, default=NORMALIZATION_SCALE, help="Scale for smoothing."
)
@click.argument("input_image", type=click.STRING)
def contrast_normalization(input_image, scale, output):
    """Automatic local contrast normalization."""

    # open the images
    im = nibabel.load(input_image)

    # load the image data and the coil correction data and apply
    data = im.get_data().astype(np.float32)
    f, w, sigma = global_scale(data)
    lc = local_contrast_normalization(f, w, sigma, scale=scale)

    # write out the result as floating point and preserve the header
    out_im = type(im)(lc.astype(np.float32), affine=None, header=im.header)
    out_im.set_data_dtype(np.float32)
    out_im.to_filename(output)

    click.echo("Wrote coil contrast normalized image to {}.".format(output))


@click.command()
@click.option(
    "--output",
    type=click.STRING,
    default="out.nii",
    help="Output filename for the corrected image.",
)
@click.option(
    "--scale", type=click.INT, default=NORMALIZATION_SCALE, help="Scale for smoothing."
)
@click.argument("input_image", type=click.STRING)
def coil_correction(input_image, scale, output):
    """Compute and apply receive coil intensity correction."""

    # open the images
    im = nibabel.load(input_image)

    # load the image data and the coil correction data and apply
    data = im.get_data().astype(np.float32)
    f, w, sigma = global_scale(data)
    data_corr = local_scale_normalization(
        f, normalization_scale=scale, w=w, sigma=sigma
    )

    # write out the result as floating point and preserve the header
    out_im = type(im)(data_corr.astype(np.float32), affine=None, header=im.header)
    out_im.set_data_dtype(np.float32)
    out_im.to_filename(output)

    click.echo("Wrote coil intensity corrected image to {}.".format(output))


@click.command(name=laplacian_pyramid)
@click.option(
    "--output",
    type=click.STRING,
    default="out.nii",
    help="Output filename for the laplacian pyramid image.",
)
@click.argument("input_image", type=click.STRING)
@click.option(
    "--num_scales", type=click.INT, default=NUM_SCALES, help="number of spatial scales"
)
@click.option(
    "--normalization_scale",
    type=click.INT,
    default=NORMALIZATION_SCALE,
    help="scale for input gain control",
)
def compute_laplacian_pyramid(input_image, num_scales, normalization_scale, output):
    """Compute rotationally invariant features."""

    click.echo(f"Compute rotationally invariant features for {input_image}.")

    # open the images
    im = nibabel.load(input_image)
    data = im.get_data().astype(np.float32)

    # Global scaling, signal likelihood, noise level
    f, w, sigma = global_scale(data)

    # Local scale normalization
    f_c = local_scale_normalization(
        f, w=w, sigma=sigma, normalization_scale=normalization_scale
    )

    # compute the laplacian pyramid
    lpyr = laplacian_pyramid(f_c, w=w, sigma=sigma, num_scales=num_scales)

    # Smush into a single array and convert to single precision
    # with pyramed level as the fourth dimension
    lpyr = np.stack(lpyr, axis=-1).astype(np.float32)

    # write out the result in the same format and preserve the header
    out_im = type(im)(lpyr, affine=None, header=im.header)
    out_im.set_data_dtype(np.float32)
    out_im.to_filename(output)

    click.echo(f"Wrote Laplacian Pyramid to {output}.")


@click.command()
@click.option(
    "--output",
    type=click.STRING,
    default="out.nii",
    help="Output filename for the features image.",
)
@click.argument("input_image", type=click.STRING)
@click.option(
    "--num_scales", type=click.INT, default=NUM_SCALES, help="number of spatial scales"
)
@click.option(
    "--normalization_scale",
    type=click.INT,
    default=NORMALIZATION_SCALE,
    help="scale for input gain control",
)
@click.option("--lowpass/--no-lowpass", default=False)
def riff(input_image, num_scales, normalization_scale, output, lowpass):
    """Compute rotationally invariant bandpass features."""

    click.echo(f"Compute rotationally invariant features for {input_image}.")

    # open the images
    im = nibabel.load(input_image)
    data = im.get_data().astype(np.float32)

    # Global scaling, signal likelihood, noise level
    f, w, sigma = global_scale(data)

    # Local scale normalization
    f = local_scale_normalization(
        f, w=w, sigma=sigma, normalization_scale=normalization_scale
    )

    # compute the frond end features
    fef = front_end_features(f, w=w, sigma=sigma, num_scales=num_scales)

    # rotational invariants
    fef = fef_rotational_invariants(fef, inplace=True)

    # convert to single precision and smush into a single array
    # with feature number as the fourth dimension
    # level changes fastest on disk, then subband, then order
    # HP,Zero,LP,|HP|,|Zero|,|LP|,|Two|,I1(Two),I2(Two)

    feats = []
    feats.append(fef["high_pass"][0].astype(np.float32))
    for n in range(num_scales):
        feats.append(fef["zero"][n][0].astype(np.float32))
    if lowpass:
        feats.append(fef["low_pass"][0].astype(np.float32))
    feats.append(fef["high_pass"][1].astype(np.float32))
    for n in range(num_scales):
        feats.append(fef["zero"][n][1].astype(np.float32))

    for n in range(num_scales):
        feats.append(fef["one"][n][0].astype(np.float32))

    M = len(fef["two"][0])
    for m in range(M):
        for n in range(num_scales):
            feats.append(fef["two"][n][m].astype(np.float32))

    feats = np.stack(feats, axis=-1)

    # write out the result in the same format and preserve the header
    out_im = type(im)(feats, affine=None, header=im.header)
    out_im.set_data_dtype(np.float32)
    out_im.to_filename(output)

    click.echo(f"Wrote rotationally invariant bandpass features to {output}.")


@click.command()
@click.option(
    "--output",
    type=click.STRING,
    default="out.nii",
    help="Output filename for the features image.",
)
@click.argument("input_image", type=click.STRING)
@click.option(
    "--num_scales", type=click.INT, default=NUM_SCALES, help="number of spatial scales"
)
@click.option(
    "--normalization_scale",
    type=click.INT,
    default=NORMALIZATION_SCALE,
    help="scale for input gain control",
)
def compute_features(input_image, num_scales, normalization_scale, output):
    """Compute rotationally invariant features."""

    click.echo(f"Compute rotationally invariant features for {input_image}.")

    # open the images
    im = nibabel.load(input_image)
    data = im.get_data().astype(np.float32)

    # Global scaling, signal likelihood, noise level
    f, w, sigma = global_scale(data)

    # Local scale normalization
    f = local_scale_normalization(
        f, w=w, sigma=sigma, normalization_scale=normalization_scale
    )

    # compute features
    feats = gaussian_pyramid_features(f, num_scales=num_scales, w=w, sigma=sigma)
    feats = np.stack(feats, axis=-1)

    # write out the result in the same format and preserve the header
    out_im = type(im)(feats, affine=None, header=im.header)
    out_im.set_data_dtype(np.float32)
    out_im.to_filename(output)



# Utility functions
import numpy as np


def select(data, mask=None):

    if isinstance(data, list):
        h = []
        for dsub in data:
            h.append(select(dsub, mask))
        return np.hstack(h)
    else:
        if mask is not None:
            if len(data.shape) == 3:
                return data.reshape(-1, 1)[mask.flatten(), :]
            else:
                return data.reshape(-1, data.shape[-1])[mask.flatten(), :]
        else:
            if len(data.shape) == 3:
                return data.reshape(-1, 1)
            else:
                return data.reshape(-1, data.shape[-1])



# Note:  This implementation of the Shepp-Logan phantom was copied from the
# from the ISMRMRD python tools on Dec. 22, 2018
# https://github.com/ismrmrd/ismrmrd-python-tools

import numpy as np


def phantom(matrix_size=256, phantom_type="Modified Shepp-Logan", ellipses=None):
    """
    Create a Shepp-Logan or modified Shepp-Logan phantom::
        phantom (n = 256, phantom_type = 'Modified Shepp-Logan', ellipses = None)
    :param matrix_size: size of imaging matrix in pixels (default 256)
    :param phantom_type: The type of phantom to produce.
        Either "Modified Shepp-Logan" or "Shepp-Logan". This is overridden
        if ``ellipses`` is also specified.
    :param ellipses: Custom set of ellipses to use.  These should be in
        the form::
            [[I, a, b, x0, y0, phi],
            [I, a, b, x0, y0, phi],
                            ...]
        where each row defines an ellipse.
        :I: Additive intensity of the ellipse.
        :a: Length of the major axis.
        :b: Length of the minor axis.
        :x0: Horizontal offset of the centre of the ellipse.
        :y0: Vertical offset of the centre of the ellipse.
        :phi: Counterclockwise rotation of the ellipse in degrees,
            measured as the angle between the horizontal axis and
            the ellipse major axis.
    The image bounding box in the algorithm is ``[-1, -1], [1, 1]``,
    so the values of ``a``, ``b``, ``x0``, ``y0`` should all be specified with
    respect to this box.
    :returns: Phantom image
    References:
    Shepp, L. A.; Logan, B. F.; Reconstructing Interior Head Tissue
    from X-Ray Transmissions, IEEE Transactions on Nuclear Science,
    Feb. 1974, p. 232.
    Toft, P.; "The Radon Transform - Theory and Implementation",
    Ph.D. thesis, Department of Mathematical Modelling, Technical
    University of Denmark, June 1996.
    """

    if ellipses is None:
        ellipses = _select_phantom(phantom_type)
    elif np.size(ellipses, 1) != 6:
        raise AssertionError("Wrong number of columns in user phantom")

    ph = np.zeros((matrix_size, matrix_size), dtype=np.float32)

    # Create the pixel grid
    ygrid, xgrid = np.mgrid[-1 : 1 : (1j * matrix_size), -1 : 1 : (1j * matrix_size)]

    for ellip in ellipses:
        intensity = ellip[0]
        a2 = ellip[1] ** 2
        b2 = ellip[2] ** 2
        x0 = ellip[3]
        y0 = ellip[4]
        phi = ellip[5] * np.pi / 180  # Rotation angle in radians

        # Create the offset x and y values for the grid
        x = xgrid - x0
        y = ygrid - y0

        cos_p = np.cos(phi)
        sin_p = np.sin(phi)

        # Find the pixels within the ellipse
        locs = (
            ((x * cos_p + y * sin_p) ** 2) / a2 + ((y * cos_p - x * sin_p) ** 2) / b2
        ) <= 1

        # Add the ellipse intensity to those pixels
        ph[locs] += intensity

    return ph


def _select_phantom(name):
    if name.lower() == "shepp-logan":
        e = _shepp_logan()
    elif name.lower() == "modified shepp-logan":
        e = _mod_shepp_logan()
    else:
        raise ValueError("Unknown phantom type: %s" % name)
    return e


def _shepp_logan():
    #  Standard head phantom, taken from Shepp & Logan
    return [
        [2, 0.69, 0.92, 0, 0, 0],
        [-0.98, 0.6624, 0.8740, 0, -0.0184, 0],
        [-0.02, 0.1100, 0.3100, 0.22, 0, -18],
        [-0.02, 0.1600, 0.4100, -0.22, 0, 18],
        [0.01, 0.2100, 0.2500, 0, 0.35, 0],
        [0.01, 0.0460, 0.0460, 0, 0.1, 0],
        [0.02, 0.0460, 0.0460, 0, -0.1, 0],
        [0.01, 0.0460, 0.0230, -0.08, -0.605, 0],
        [0.01, 0.0230, 0.0230, 0, -0.606, 0],
        [0.01, 0.0230, 0.0460, 0.06, -0.605, 0],
    ]


def _mod_shepp_logan():
    #  Modified version of Shepp & Logan's head phantom,
    #  adjusted to improve contrast.  Taken from Toft.
    return [
        [1, 0.69, 0.92, 0, 0, 0],
        [-0.80, 0.6624, 0.8740, 0, -0.0184, 0],
        [-0.20, 0.1100, 0.3100, 0.22, 0, -18],
        [-0.20, 0.1600, 0.4100, -0.22, 0, 18],
        [0.10, 0.2100, 0.2500, 0, 0.35, 0],
        [0.10, 0.0460, 0.0460, 0, 0.1, 0],
        [0.10, 0.0460, 0.0460, 0, -0.1, 0],
        [0.10, 0.0460, 0.0230, -0.08, -0.605, 0],
        [0.10, 0.0230, 0.0230, 0, -0.606, 0],
        [0.10, 0.0230, 0.0460, 0.06, -0.605, 0],
    ]



import logging


logger = logging.getLogger(__name__)

# To show debug messages, uncomment this
# from outside the module, import the logger object and use something like this
# logger.setLevel(logging.DEBUG)

# add a console handler
ch = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)



import numpy as np

SIGNAL_MAX_ITER = 5
SIGMA_SCALE = 0.01


def stats(data, w=None):
    """Weighted mean and mean and absolute deviation

    mean = sum(w*data)/sum(w)
    sigma = sum(w*|data-m|)/sum(w)
    """
    if w is not None:
        mu = np.sum(w * data) / np.sum(w)
        sigma = np.sum(w * np.abs(data - mu)) / np.sum(w)
    else:
        mu = np.mean(data)
        sigma = np.mean(np.abs(data - mu))

    return mu, sigma


def signal_weight(data, mu=1.0):
    """Return a weight for the signal

    w(x) = 1 - exp(-x/mu)
    """
    w = 1.0 - np.exp(-data / mu)
    return w


def noise_weight(data, mu=1.0):
    """Return a weight for the noise

    w(x) = (x>0) * exp(-x/mu)
    """
    w = (data > 0) * np.exp(-data / mu)
    return w


def signal_stats(data, niter=SIGNAL_MAX_ITER):
    """Iterative weighted estimation of signal mean and signal lower bound"""

    d = data[data > 0]

    # Assume that everything is signal to start with
    mu_s = np.mean(d)

    # iterate (two or three iterations is enough)
    for _iter in range(niter):
        # Get the likelihood
        w = signal_weight(d, mu_s)
        # Update
        mu_s = np.sum(w * d) / np.sum(w)

    # Estimate a lower bound on the signal
    w_n = noise_weight(d, mu_s)
    mu_n = np.sum(d * w_n) / np.sum(w_n)

    return mu_s, mu_n


def global_scale(data, sigma_scale=SIGMA_SCALE, niter=SIGNAL_MAX_ITER):
    """Estimate signal and noise statistics and apply a global scale

    :param data: 2D or 3D magnitude image
    :return: scaled data (f)
             signal likelihood (w_s)
             noise level relative to signal (sigma_n)
    """

    # Estimate the signal mean and signal lower bound
    mu_s, mu_n = signal_stats(data, niter)

    # Compute a signal weight based on the signal lower bound
    w = signal_weight(data, mu_n)

    # Scale the signal by the signal mean
    f = data / mu_s

    # Set the reguralization factor to 0.01*scaled signal lower bound
    sigma = sigma_scale * mu_n / mu_s

    return f, w, sigma



import numpy as np

from .filters import (
    _pinv,
    dog,
    dog_rotational_invariants,
    gradient_amplitude,
    hessian_amplitude,
    hessian_trace,
)

# Number of spatial scales
NUM_SCALES = 4

# Scales for normalization / gain control
NORMALIZATION_SCALE = 5


def rectify(data, polarity=1):
    """
    Rectify the data, polarity is either 1 or -1
    """
    if polarity not in [1, -1]:
        raise RuntimeError("Rectification polarity is either 1 or -1.")

    output = polarity * data * ((polarity * data) > 0)

    return output


def local_scale_normalization(
    d, normalization_scale=NORMALIZATION_SCALE, w=None, sigma=0.001
):
    """
    Local normalization by the signal mean.
    """
    f = d * _pinv(
        dog(d, order=0, scale=normalization_scale, w=w, sigma=sigma), sigma=sigma
    )

    return f


def local_contrast_normalization(
    d, normalization_scale=NORMALIZATION_SCALE, w=None, sigma=0.001
):
    """
    Local contrast normalization
    Subtract the local mean and scale by the local mean absolute deviation
    """
    dev = d - dog(d, order=0, scale=normalization_scale, w=w, sigma=sigma)
    lc = dev * _pinv(
        dog(np.abs(dev), order=0, scale=normalization_scale, w=w, sigma=sigma),
        sigma=sigma,
    )

    return lc


def gaussian_pyramid(d, order=0, num_scales=NUM_SCALES, w=None, sigma=0.001):
    """Pyramid from gaussian derivative filters with signal likelihood weighting

    :param d: 2D or 3D numpy array
    :param order: Gaussian derivative order, 0, 1, or 2
    :param w: signal likelihood
    :param sigma: noise level
    :param nscales: int number of scales
    :return: list

    The pyramid is organized by scale,
    Level 0, Level 1, .., Level num_scales-1
    """

    pyr = [
        dog(d, order=order, scale=scale, w=w, sigma=sigma)
        for scale in range(num_scales)
    ]

    return pyr


def difference_pyramid(pyr, inplace=False):
    """Difference between levels of a derivative of gaussian pyramid

    :param d: gaussian pyramid
    :param inplace: bool, do the work in place or make a copy, default False
    :return: list of differences

    The output has length one less than the input
    diff_pyr = [pyr[0] - pyr[1], pyr[1] - pyr[2], ..., pyr[N-2] - pyr[N-1]]
    """

    N = len(pyr)
    if type(pyr[0]) == list:
        M = len(pyr[0])
    else:
        M = 1

    if inplace:
        output = pyr
    else:
        if M > 1:
            output = [[None] * M] * N
        else:
            output = [None] * N

    # the first N-1 levels
    for n in range(N - 1):
        if M > 1:
            for m in range(M):
                output[n][m] = pyr[n][m] - pyr[n + 1][m]
        else:
            output[n] = pyr[n] - pyr[n + 1]

    # pop off the last level (the low pass)
    output = output[:-1]

    return output


def laplacian_pyramid(d, num_scales=NUM_SCALES, w=None, sigma=0.001):
    """Laplacian pyramid from gaussian filters with signal likelihood weighting

    :param d: 2D or 3D numpy array
    :param w: signal likelihood
    :param sigma: noise level
    :param nscales: int number of scales
    :return: list of length num_scales + 1

    The pyramid is organized by scale
    high pass, level 0, level 2, .., level num_scales-1, low pass

    Adding the output back together produces the original input data
    """

    # Gaussian pyramid.  Get 1 extra level.
    pyr = gaussian_pyramid(d, order=0, num_scales=num_scales + 1, w=w, sigma=sigma)

    # Save the high pass and the low pass
    high_pass = d - pyr[0]
    low_pass = pyr[-1]

    # The difference pyramid (in place)
    pyr = difference_pyramid(pyr, inplace=True)

    # Insert the high at the front and the low pass at the end
    pyr.insert(0, high_pass)
    pyr.append(low_pass)

    return pyr


def front_end_features(d, w=None, sigma=0.001, num_scales=NUM_SCALES):
    """Multi-scale image filters from the zeroth, first, and
    second order gaussian derivatives with signal likelihood weighting

    :param d: 2D or 3D numpy array
    :param w: signal likelihood
    :param sigma: noise level
    :param num_scales: int number of scales
    :return: dictionary {high_pass, zelist, one dictionary per level

    High pass
    Multi scale bandpass derivatives are organized by scale,
    zero (band pass), first (gradient), second (hessian)
    """

    feats = {
        "high_pass": None,
        "zero": [None] * num_scales,
        "one": [None] * num_scales,
        "two": [None] * num_scales,
        "low_pass": None,
    }

    # High pass, zeroth order, low pass
    # i.e. Laplacian pyramid
    pyr = laplacian_pyramid(d, num_scales=num_scales, w=w, sigma=sigma)
    feats["high_pass"] = pyr[0]
    feats["zero"] = pyr[1:-1]
    feats["low_pass"] = pyr[-1]
    del pyr

    # First order, gradient
    feats["one"] = gaussian_pyramid(
        d, order=1, num_scales=num_scales + 1, w=w, sigma=sigma
    )
    feats["one"] = difference_pyramid(feats["one"], inplace=True)

    # Second order, hessian
    feats["two"] = gaussian_pyramid(
        d, order=2, num_scales=num_scales + 1, w=w, sigma=sigma
    )
    feats["two"] = difference_pyramid(feats["two"], inplace=True)

    return feats


def fef_rotational_invariants(fef, inplace=True):
    """Rotational invariants of the front end features"""

    num_scales = len(fef["zero"])

    # Rotational invariants
    if inplace:
        rfef = fef
    else:
        rfef = {
            "high_pass": None,
            "zero": [None] * num_scales,
            "one": [None] * num_scales,
            "two": [None] * num_scales,
            "low_pass": None,
        }

    rfef["high_pass"] = dog_rotational_invariants(fef["high_pass"], order=0)
    rfef["low_pass"] = [fef["low_pass"]]
    for n in range(num_scales):
        rfef["zero"][n] = dog_rotational_invariants(fef["zero"][n], order=0)
        rfef["one"][n] = dog_rotational_invariants(fef["one"][n], order=1)
        rfef["two"][n] = dog_rotational_invariants(fef["two"][n], order=2)

    return rfef


def gaussian_pyramid_features(d, w, sigma, num_scales=NUM_SCALES):
    """Rotational invariants of the gaussian pyramid features"""

    gauss = gaussian_pyramid(d, order=0, num_scales=num_scales, w=w, sigma=sigma)
    grad = gaussian_pyramid(d, order=1, num_scales=num_scales, w=w, sigma=sigma)
    grad = [gradient_amplitude(x) for x in grad]
    hess = gaussian_pyramid(d, order=2, num_scales=num_scales, w=w, sigma=sigma)
    lap = [hessian_trace(x) for x in hess]
    norm = [hessian_amplitude(x) for x in hess]
    features = [d] + gauss + grad + lap + norm

    return features
